:- module(concert).

:- export
        none_problem/3,
        problem/3,
        gap_sbds_problem/3,
        gap_sbdd_problem/3,
        isbds_problem/3,
        static_problem/3,
        sbdd_ssb_problem/5.

:- lib(ic).
:- lib(branch_and_bound).
:- lib(ldsb).
:- lib(ldsb_syms).
:- lib(ldsb_sbds).
:- lib(ldsb_static).
:- lib(ldsb_misc).
:- lib(ic_gap_sbds).
:- lib(ic_gap_sbdd).
:- lib(ic_sbds).

% Is Finish inclusive or exclusive?
% Assumed inclusive.

setup(Offers, NHalls, Xs, Cost, Partitions) :-
        dim(Offers, [NOffers]),
        dim(Xs, [NOffers]),

        MaxVal is NHalls+1,
        Xs #:: 1..MaxVal,
        Unassigned = MaxVal,

        % Look through each pair of offers -- if they overlap, they
        % must not share a hall.
        ( for(I,1,NOffers),
          param(NOffers, Xs, Offers, Unassigned) do
          I1 is I+1,
          ( for(J,I1,NOffers),
            param(I, Xs, Offers, Unassigned) do
            arg(I, Xs, X), arg(I, Offers, OfferX),
            arg(J, Xs, Y), arg(J, Offers, OfferY),
            ( overlap(OfferX,OfferY) -> dont_share(Unassigned,X,Y) ; true ) ) ),

        % The profit is the sum of the assigned halls.
        ( foreachelem(X, Xs),
          foreachelem(Offer, Offers),
          foreach(P, Profits),
          param(Unassigned) do
            Offer = offer(_Start,_Finish,Price),
            P = (X #\= Unassigned)*Price ),
        Profit #= sum(Profits),
        Cost #= -Profit,

        % Partition the offers into identical subsets.
        ( foreachelem(O, Offers), foreachelem(X, Xs), foreach(O-X, Pairs) do true ),
        partition_offers(Pairs, PairPartitions),
        % Convert the offer-var pairs into just variables.
        ( foreach(PairPart, PairPartitions), foreach(Part, Partitions) do
          ( foreach(_O-X, PairPart), foreach(X, Part) do true ) ).

problem_symmetries(_Matrix, NHalls, Partitions, Syms) :-
        % Identical offers are interchangeable.
        ( foreach(Part, Partitions),
          foreach(VarSym, VarSyms) do
            ( foreach(X, Part), foreach(X, PartVars) do true ),
            VarSym = variable_interchange(PartVars) ),
        % Halls are interchangeable but not all values!
        ( for(I,1,NHalls), foreach(I, Vals) do true ),
        ValSym = value_interchange(Vals),
        
        Syms = [ValSym|VarSyms].

ldsb_symmetries(Matrix, NHalls, Partitions) :-
        problem_symmetries(Matrix, NHalls, Partitions, Syms),
        ldsb_initialise(Matrix, Syms).

gap_sbds_symmetries(_Matrix, NHalls, Partitions) :-
        % Find the maximum size partition.
        ( foreach(P, Partitions), foreach(Length, Lengths) do length(P,Length) ),
        eclipse_language:max(Lengths, MaxSize),
        % Make a matrix.
        length(Partitions, NParts),
        dim(NewMatrix, [NParts, MaxSize]),
        % Copy the elements of Partitions into NewMatrix.
        ( foreach(P, Partitions), foreacharg(Q, NewMatrix), foreach(Sym, Syms), for(I,1,NParts) do
            ( foreach(X, P), count(J,1,_), param(Q) do arg(J, Q, X) ),
            length(P,N),
            Sym = symmetry(s_n, [position:1..N], [partition:I]) ),

        NValues is NHalls+1,
        SymList = [ symmetry(s_n, [value:1..NHalls], []) | Syms ],
        ic_gap_sbds:sbds_initialise(NewMatrix, [partition,position], value:1..NValues,
                                    SymList, []).
gap_sbdd_symmetries(_Matrix, NHalls, Partitions) :-
        % Find the maximum size partition.
        ( foreach(P, Partitions), foreach(Length, Lengths) do length(P,Length) ),
        eclipse_language:max(Lengths, MaxSize),
        % Make a matrix.
        length(Partitions, NParts),
        dim(NewMatrix, [NParts, MaxSize]),
        % Copy the elements of Partitions into NewMatrix.
        ( foreach(P, Partitions), foreacharg(Q, NewMatrix), foreach(Sym, Syms), for(I,1,NParts), param(MaxSize) do
            length(P,N),
            ( foreach(X, P), count(J,1,N), param(Q) do arg(J, Q, X) ),
            N1 is N+1,
            % SBDD needs dummy variables here (for some reason).
            % (The whole matrix must have proper IC variables in it.)
            ( for(J, N1, MaxSize), param(Q) do arg(J,Q,Y), Y #:: 0..1),
            Sym = symmetry(s_n, [position:1..N], [partition:I]) ),

        NValues is NHalls+1,
        SymList = [ symmetry(s_n, [value:1..NHalls], []) | Syms ],
        ic_gap_sbdd:sbdd_initialise(NewMatrix, [partition,position], value:1..NValues,
                                    SymList, []).

isbds_symmetries(Matrix, NHalls, Partitions) :-
        problem_symmetries(Matrix, NHalls, Partitions, Syms),
        convert_symmetries(pairs, Matrix, Syms, SBDSSyms),
        ic_sbds:sbds_initialise(Matrix, 1, SBDSSyms, #=, []).

partition_offers([], []).
partition_offers([O-X|Rest], [ThisPart|OtherParts]) :-
        filter_offers(O-X, Rest, Same, Different),
        ThisPart = [O-X|Same],
        partition_offers(Different, OtherParts).
filter_offers(_, [], [], []).
filter_offers(O-X, [O2-X2|Rest], Same, Different) :-
        filter_offers(O-X, Rest, Same0, Diff0),
        ( O == O2 -> Same = [O2-X2|Same0], Different = Diff0
        ; Same = Same0, Different = [O2-X2|Diff0] ).

overlap(offer(S1,F1,_), offer(S2,F2,_)) :-
        once ( between(S1, S2, F2) ;
               between(S2, S1, F1) ).

between(X, L, R) :- L =< X, X =< R.

dont_share(Unassigned, X, Y) :-
        (X #= Unassigned) or (Y #= Unassigned) or (X #\= Y).
            
instance_dir('siglex-data/concert').

read_instance(NOffers, InstanceNumber, Offers) :-
        instance_dir(Dir),
        sprintf(Filename, "concert.%d.8.100.%d", [NOffers, InstanceNumber]),
        join_string([Dir, Filename], "/", Path),

        open(Path, read, Stream),
        % Number of offers.
        read_token(Stream, NOffers, integer),
        dim(Offers, [NOffers]),

        % List of offers.
        ( foreachelem(offer(Start, Finish, Price), Offers),
          for(_I,1,NOffers), param(Stream) do
            read_token(Stream, Start, integer),
            read_token(Stream, Finish, integer),
            read_token(Stream, Price, integer) ).

% solve(NOffers, InstanceNumber) :-
%         read_instance(NOffers, InstanceNumber, Os),
%         setup(Os, 8, Xs, Profit),

%         Cost #= (-Profit),
%         bb_min(ldsb_labeling(Xs), Cost, _).


none_problem(concert(NOffers, Instance), Matrix, Cost) :-
        NHalls = 10,
        read_instance(NOffers, Instance, Offers), setup(Offers, NHalls, Matrix, Cost, VarParts).
problem(concert(NOffers, Instance), Matrix, Cost) :-
        NHalls = 10,
        read_instance(NOffers, Instance, Offers), setup(Offers, NHalls, Matrix, Cost, VarParts),
        ldsb_symmetries(Matrix, NHalls, VarParts).
gap_sbds_problem(concert(NOffers, Instance), Matrix, Cost) :-
        NHalls = 10,
        read_instance(NOffers, Instance, Offers), setup(Offers, NHalls, Matrix, Cost, VarParts),
        gap_sbds_symmetries(Matrix, NHalls, VarParts).
gap_sbdd_problem(concert(NOffers, Instance), Matrix, Cost) :-
        NHalls = 10,
        read_instance(NOffers, Instance, Offers), setup(Offers, NHalls, Matrix, Cost, VarParts),
        gap_sbdd_symmetries(Matrix, NHalls, VarParts).
isbds_problem(concert(NOffers, Instance), Matrix, Cost) :-
        NHalls = 10,
        read_instance(NOffers, Instance, Offers), setup(Offers, NHalls, Matrix, Cost, VarParts),
        isbds_symmetries(Matrix, NHalls, VarParts).
static_problem(concert(NOffers, Instance), Matrix, Cost) :-
        NHalls = 10,
        read_instance(NOffers, Instance, Offers), setup(Offers, NHalls, Matrix, Cost, VarParts),
        problem_symmetries(Matrix, NHalls, VarParts, Syms),
        post_symmetries(Matrix, Syms).
sbdd_ssb_problem(concert(NOffers, Instance), Vector, Cost, VarPart, ValPart) :-
        NHalls = 10,
        read_instance(NOffers, Instance, Offers), setup(Offers, NHalls, Matrix, Cost, VarParts),
        problem_symmetries(Matrix, NHalls, VarParts, Syms),
        symmetries_to_ssb(Matrix, Syms, Vector, VarPart, ValPart).
