% ECLiPSe SAMPLE CODE
%
% AUTHOR:	Joachim Schimpf, IC-Parc
%               (but changed a bit by Chris Mears)
%
% DESCRIPTION:	A magic square of size N is an NxN square filled
%		with the numbers from 1 to N^2 such that the sums of
%		each row, each column and the two main diagonals are
%		equal.
%
%		Some interesting goals:
%
%		magic(4),fail.	% all 880 solutions
%		magic(5).
%		magic(7).

:- module(magicsquare).

:- export
        magicsquare/2,
        none_problem/2,
        problem/2,
        problem_symmetries/2,
        gap_sbds_problem/2,
        gap_sbdd_problem/2,
        isbds_problem/2,
        static_problem/2,
        sbdd_ssb_problem/4.

:- lib(ic).
:- lib(ic_sbds).
:- lib(ic_gap_sbds).
:- lib(ic_gap_sbdd).

:- lib(ldsb).
:- lib(ldsb_syms).
:- lib(ldsb_sbds).
:- lib(ldsb_static).
:- lib(ldsb_misc).

magicsquare(N, Square) :-
	NN is N*N,				% number of square fields
	Sum is N*(NN+1)//2,			% this is the magical sum

	dim(Square, [N,N]),			% make the variables
	Square[1..N,1..N] :: 1..NN,		% their range
	Rows is Square[1..N,1..N],
	flatten(Rows, Vars),
	alldifferent(Vars),			% they are all different

	(
	    for(I,1,N),
	    foreach(U,UpDiag),
	    foreach(D,DownDiag),
	    param(N,Square,Sum)
	do
	    Sum #= sum(Square[I,1..N]),		% sum of row I
	    Sum #= sum(Square[1..N,I]),		% sum of column I
	    U is Square[I,I],
	    D is Square[I,N+1-I]
	),
	Sum #= sum(UpDiag),			% diagonal sums
	Sum #= sum(DownDiag).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

problem_symmetries(Matrix, Syms) :-
        rows_reflect(Matrix, RowSym),
        columns_reflect(Matrix, ColSym),
        diagonal_reflect(Matrix, DiagSym),
        values_reflect(Matrix, ValSym),
        Syms = [RowSym, ColSym, DiagSym, ValSym].

problem(N, Matrix) :-
        magicsquare(N, Matrix),
        problem_symmetries(Matrix, Syms),
        ldsb_initialise(Matrix, Syms).

none_problem(N, Matrix) :- magicsquare(N, Matrix).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

gap_sbds_problem(N,M) :-
        magicsquare(N,M),
        NN is N*N,
        ic_gap_sbds:sbds_initialise(M, [rows, cols], values:1..NN,
                        [symmetry(d_4, [rows, cols], [])],
                        []).

gap_sbdd_problem(N,M) :-
        magicsquare(N,M),
        NN is N*N,
        ic_gap_sbdd:sbdd_initialise(M, [rows, cols], values:1..NN,
                        [symmetry(d_4, [rows, cols], [])],
                        []).

isbds_problem(N, M) :-
        magicsquare(N,M),
        problem_symmetries(M, Syms),
        convert_symmetries(pairs, M, Syms, SBDSSyms),
        ic_sbds:sbds_initialise(M, 2, SBDSSyms, #=, []).

static_problem(N, Matrix) :-
        magicsquare(N, Matrix),
        problem_symmetries(Matrix, Syms),
        post_symmetries(Matrix, Syms).

sbdd_ssb_problem(N, Vector, VarPart, ValPart) :-
        magicsquare(N, Matrix),
        problem_symmetries(Matrix, Syms),
        symmetries_to_ssb(Matrix, Syms, Vector, VarPart, ValPart).
